from functools import partial
from pathlib import Path
import tempfile
from typing import Final, Generator

import pytest

from ohmqtt.mqtt_spec import MQTTQoS
from ohmqtt.packet import MQTTPublishPacket, MQTTPubRelPacket
from ohmqtt.persistence.base import Persistence
from ohmqtt.persistence.in_memory import InMemoryPersistence
from ohmqtt.persistence.sqlite import SQLitePersistence
from ohmqtt.property import MQTTPublishProps
from ohmqtt.topic_alias import AliasPolicy


SQLiteInMemory = partial(SQLitePersistence, ":memory:", db_fast=True)


@pytest.fixture
def tempdbpath() -> Generator[str, None, None]:
    """Fixture to create a temporary directory for testing."""
    with tempfile.TemporaryDirectory() as tempdir:
        yield str(Path(tempdir) / "test.db")


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_open(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()

    persistence.open("test_client")
    persistence.add(
        "test/topic",
        b"test payload",
        qos=MQTTQoS.Q1,
        retain=False,
        properties=MQTTPublishProps(ResponseTopic="response/topic"),
        alias_policy=AliasPolicy.TRY,
    )
    assert len(persistence) == 1
    mids = persistence.get(1)
    rendered = persistence.render(mids[0])
    assert isinstance(rendered.packet, MQTTPublishPacket)
    assert not rendered.packet.dup

    persistence.open("test_client")
    assert len(persistence) == 1
    rendered.packet.dup = True
    mids = persistence.get(1)
    assert rendered.packet == persistence.render(mids[0]).packet

    persistence.open("test_client2")
    assert len(persistence) == 0


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_happy_path_qos1(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")
    assert len(persistence) == 0

    # Add a message to the store.
    handle = persistence.add(
        "test/topic",
        b"test payload",
        qos=MQTTQoS.Q1,
        retain=False,
        properties=MQTTPublishProps(ResponseTopic="response/topic"),
        alias_policy=AliasPolicy.TRY,
    )
    assert len(persistence) == 1

    # Retrieve the PUBLISH from the store.
    message_ids = persistence.get(10)
    assert len(message_ids) == 1
    expected_packet = MQTTPublishPacket(
        packet_id=1,
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q1,
        retain=False,
        properties=MQTTPublishProps(ResponseTopic="response/topic"),
    )

    # Render the message, marking it as inflight.
    rendered = persistence.render(message_ids[0])
    assert rendered.packet == expected_packet
    assert rendered.alias_policy == AliasPolicy.TRY
    assert len(persistence) == 1

    # We should not be able to retrieve the message again.
    assert len(persistence.get(10)) == 0

    # Acknowledge the message.
    persistence.ack(1)
    assert len(persistence) == 0
    assert handle.is_acked()


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_happy_path_qos2(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")
    assert len(persistence) == 0

    # Add a message to the store.
    handle = persistence.add(
        "test/topic",
        b"test payload",
        qos=MQTTQoS.Q2,
        retain=False,
        properties=MQTTPublishProps(ResponseTopic="response/topic"),
        alias_policy=AliasPolicy.TRY,
    )
    assert len(persistence) == 1

    # Retrieve the PUBLISH from the store.
    message_ids = persistence.get(10)
    assert len(message_ids) == 1
    expected_publish = MQTTPublishPacket(
        packet_id=1,
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q2,
        retain=False,
        properties=MQTTPublishProps(ResponseTopic="response/topic"),
    )

    # Render the message, marking it as inflight.
    rendered = persistence.render(message_ids[0])
    assert rendered.packet == expected_publish
    assert rendered.alias_policy == AliasPolicy.TRY
    assert len(persistence) == 1

    # We should not be able to retrieve the message again.
    assert len(persistence.get(10)) == 0

    # Acknowledge the message.
    persistence.ack(1)
    assert len(persistence) == 1
    assert not handle.is_acked()

    # Retrieve the PUBREL from the store.
    message_ids = persistence.get(10)
    assert len(message_ids) == 1
    expected_pubrel = MQTTPubRelPacket(packet_id=1)

    # Render the message, marking it as inflight.
    rendered = persistence.render(message_ids[0])
    assert rendered.packet == expected_pubrel
    assert rendered.alias_policy == AliasPolicy.NEVER
    assert len(persistence) == 1

    # We should not be able to retrieve the message again.
    assert len(persistence.get(10)) == 0

    # Acknowledge the message.
    persistence.ack(1)
    assert len(persistence) == 0
    assert handle.is_acked()


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_properties(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")

    # Add a message with all the properties.
    packet = MQTTPublishPacket(
        packet_id=1,
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q1,
        retain=False,
        properties=MQTTPublishProps(
            ResponseTopic="response/topic",
            CorrelationData=b"correlation data",
            MessageExpiryInterval=60,
            UserProperty=[("key", "value")],
            SubscriptionIdentifier={123},
        ),
    )
    persistence.add(
        topic=packet.topic,
        payload=packet.payload,
        qos=packet.qos,
        retain=packet.retain,
        properties=packet.properties,
        alias_policy=AliasPolicy.TRY,
    )
    assert len(persistence) == 1

    # Retrieve the PUBLISH from the store.
    message_ids = persistence.get(10)
    rendered = persistence.render(message_ids[0])
    assert rendered.packet == packet


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_loose_alias(persistence_class: type[Persistence]) -> None:
    """We should not be able to add a message with an alias policy of ALWAYS."""
    persistence = persistence_class()
    persistence.open("test_client")

    # Add a message with all the properties.
    packet = MQTTPublishPacket(
        packet_id=1,
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q1,
        retain=False,
        properties=MQTTPublishProps(),
    )
    with pytest.raises(ValueError):
        persistence.add(
            topic=packet.topic,
            payload=packet.payload,
            qos=packet.qos,
            retain=packet.retain,
            properties=packet.properties,
            alias_policy=AliasPolicy.ALWAYS,
        )
    assert len(persistence) == 0


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_incoming_qos2(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")

    pub_packet = MQTTPublishPacket(
        packet_id=1,
        qos=MQTTQoS.Q1,
    )
    with pytest.raises(ValueError):
        persistence.check_rec(pub_packet)
    with pytest.raises(ValueError):
        persistence.set_rec(pub_packet)

    pub_packet.qos = MQTTQoS.Q2

    assert persistence.check_rec(pub_packet) is True
    persistence.set_rec(pub_packet)
    assert persistence.check_rec(pub_packet) is False

    rel_packet = MQTTPubRelPacket(packet_id=1)
    persistence.rel(rel_packet)

    assert persistence.check_rec(pub_packet) is True


@pytest.mark.parametrize("qos", [MQTTQoS.Q1, MQTTQoS.Q2])
@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_queue(qos: MQTTQoS, persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")

    rng = range(1, 3)
    for n in rng:
        persistence.add(
            topic=str(n),
            payload=b"bar",
            qos=qos,
            retain=False,
            properties=MQTTPublishProps(),
            alias_policy=AliasPolicy.NEVER,
        )
    for n in rng:
        queued = persistence.get(3)
        rendered = persistence.render(queued[0])
        assert isinstance(rendered.packet, MQTTPublishPacket)
        assert rendered.packet.topic == str(n)
        assert rendered.packet.packet_id == n


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_render_order(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")

    for _ in (1, 2):
        persistence.add(
            topic="foo",
            payload=b"bar",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(),
            alias_policy=AliasPolicy.NEVER,
        )
    queued = persistence.get(3)
    with pytest.raises(ValueError):
        persistence.render(queued[1])


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_unknown_ack(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")

    with pytest.raises(ValueError, match="Unknown packet_id: 42"):
        persistence.ack(42)


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_unknown_render(persistence_class: type[Persistence]) -> None:
    persistence = persistence_class()
    persistence.open("test_client")

    with pytest.raises(KeyError):
        persistence.render(42)


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_unlimited_ids(persistence_class: type[Persistence]) -> None:
    "Should be able to queue unlimited messages."
    persistence = persistence_class()
    persistence.open("test_client")

    blocks: Final = 3  # How many blocks of 65535 packets to try
    num: Final = 65535 * blocks  # Total packets
    for _ in range(num):
        persistence.add("foo", b"bar", MQTTQoS.Q1, False, MQTTPublishProps(), AliasPolicy.NEVER)

    checked = 0
    for block in range(blocks):
        for n in range(1, 65536):
            checked += 1
            try:
                pending = persistence.get(1)
                rendered = persistence.render(pending[0])
                assert isinstance(rendered.packet, MQTTPublishPacket)
                assert rendered.packet.packet_id == n
            except Exception:
                print(f"{block=} {n=}")
                raise
            persistence.ack(rendered.packet.packet_id)
    assert checked == num


@pytest.mark.parametrize("db_fast", [True, False])
def test_persistence_sqlite_open(db_fast: bool, tempdbpath: str) -> None:
    """Test the SQLitePersistence class with a resume scenario."""
    persistence = SQLitePersistence(tempdbpath, db_fast=db_fast)
    persistence.open("test_client")
    assert len(persistence) == 0

    # Add a message to the store.
    outgoing_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q1,
        retain=False,
        properties=MQTTPublishProps(ResponseTopic="response/topic"),
    )
    persistence.add(
        topic=outgoing_packet.topic,
        payload=outgoing_packet.payload,
        qos=outgoing_packet.qos,
        retain=outgoing_packet.retain,
        properties=outgoing_packet.properties,
        alias_policy=AliasPolicy.TRY,
    )
    assert len(persistence) == 1

    # Mark the message inflight by rendering it.
    persistence.render(1)
    assert len(persistence.get(1)) == 0

    # Simulate an incoming QoS 2 message.
    incoming_packet = MQTTPublishPacket(packet_id=1, qos=MQTTQoS.Q2)
    persistence.set_rec(incoming_packet)

    # Close and reopen the persistence store.
    persistence = SQLitePersistence(tempdbpath)
    # This should not clear the store.
    persistence.open("test_client")
    assert len(persistence) == 1

    # We should be able to retrieve the message again.
    message_ids = persistence.get(1)
    assert len(message_ids) == 1
    # When rendering a second time, the packet should have the dup flag set.
    outgoing_packet.dup = True
    outgoing_packet.packet_id = 1
    rendered = persistence.render(message_ids[0])
    assert rendered.packet == outgoing_packet
    assert rendered.alias_policy == AliasPolicy.TRY

    # We should filter a duplicate incoming QoS 2 message.
    assert persistence.check_rec(incoming_packet) is False

    # Now open with a different client ID.
    persistence = SQLitePersistence(tempdbpath)
    # This should clear the store.
    persistence.open("test_client_2")
    assert len(persistence) == 0
    assert persistence.check_rec(incoming_packet) is True



from typing import cast
from unittest.mock import Mock, MagicMock
import weakref

import pytest
from pytest_mock import MockerFixture

from ohmqtt.client import Client
from ohmqtt.connection import Connection, MessageHandlers, InvalidStateError
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import (
    MQTTPublishPacket,
    MQTTSubscribePacket,
    MQTTUnsubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
)
from ohmqtt.property import MQTTPublishProps, MQTTSubscribeProps
from ohmqtt.subscriptions import Subscriptions, RetainPolicy


def dummy_callback(client: Client, packet: MQTTPublishPacket) -> None:
    """Dummy SubscribeCallback for testing."""


@pytest.fixture
def mock_client(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Client)  # type: ignore[no-any-return]


@pytest.fixture
def mock_connection(mocker: MockerFixture) -> Mock:
    conn = mocker.MagicMock(spec=Connection)
    conn.fsm.cond.__enter__.return_value = mocker.Mock()
    conn.fsm.lock.__enter__.return_value = mocker.Mock()
    return conn  # type: ignore[no-any-return]


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> MagicMock:
    return mocker.MagicMock(spec=MessageHandlers)  # type: ignore[no-any-return]


def test_subscriptions_registration(mock_client: Mock, mock_connection: Mock) -> None:
    handlers = MessageHandlers()
    with handlers:
        subscriptions = Subscriptions(handlers, mock_connection, weakref.ref(mock_client))
    assert subscriptions.handle_suback in handlers.get_handlers(MQTTSubAckPacket)
    assert subscriptions.handle_unsuback in handlers.get_handlers(MQTTUnsubAckPacket)


@pytest.mark.parametrize("max_qos", [MQTTQoS.Q0, MQTTQoS.Q1, MQTTQoS.Q2])
@pytest.mark.parametrize("no_local", [True, False])
@pytest.mark.parametrize("retain_as_published", [True, False])
@pytest.mark.parametrize("retain_policy", [RetainPolicy.NEVER, RetainPolicy.ONCE, RetainPolicy.ALWAYS])
def test_subscriptions_subscribe_opts(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    max_qos: MQTTQoS,
    no_local: bool,
    retain_as_published: bool,
    retain_policy: RetainPolicy
) -> None:
    """Test subscribing with options."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe(
        "test/topic",
        dummy_callback,
        max_qos=max_qos,
        share_name="test_share",
        no_local=no_local,
        retain_as_published=retain_as_published,
        retain_policy=retain_policy,
        sub_id=23,
        user_properties=[("key", "value")],
    )
    expected_opts = max_qos.value | (retain_policy << 4) | (retain_as_published << 3) | (no_local << 2)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("$share/test_share/test/topic", expected_opts)],
        packet_id=1,
        properties=MQTTSubscribeProps(
            SubscriptionIdentifier={23},
            UserProperty=(("key", "value"),),
        ),
    ))


def test_subscriptions_multi_subscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test subscribing multiple times to the same topic before acknowledgement."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle1 = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle1 is not None

    handle2 = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle2 is not None
    assert handle1 != handle2

    assert handle1.wait_for_ack(timeout=0.001) is None
    assert handle1.ack is None
    assert handle1.failed

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    )
    subscriptions.handle_suback(suback_packet)

    assert handle2.wait_for_ack(timeout=0.1) == suback_packet
    assert handle2.ack == suback_packet


def test_subscriptions_resubscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
) -> None:
    """Test reconfiguring an existing and acknowledged subscription."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback, no_local=True)
    expected_opts = 2 | (True << 2)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", expected_opts)],
        packet_id=1,
    ))
    mock_connection.reset_mock()
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    subscriptions.subscribe("test/topic", dummy_callback, no_local=False)
    expected_opts = 2
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", expected_opts)],
        packet_id=2,
    ))
    mock_connection.reset_mock()
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=2,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))


def test_subscriptions_fast_subscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test subscribing and unsubscribing quickly."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.unsubscribe("test/topic")
    mock_connection.send.assert_called_once_with(MQTTUnsubscribePacket(
        topics=["test/topic"],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    # We should not send another UNSUBSCRIBE packet after the SUBACK
    subscriptions.unsubscribe("test/topic")
    mock_connection.send.assert_not_called()

    subscriptions.subscribe("test/topic", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=2,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_unsuback(MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.Success],
    ))

    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=2,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))


def test_subscriptions_wait_for_suback(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test waiting for a SUBACK packet."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle is not None
    assert handle.wait_for_ack(timeout=0.001) is None

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    )
    subscriptions.handle_suback(suback_packet)

    assert handle.wait_for_ack(timeout=0.001) == suback_packet
    assert handle.ack == suback_packet

    mock_handlers.reset_mock()
    del mock_handlers
    mock_client.reset_mock()
    del mock_client
    mock_connection.reset_mock()
    del mock_connection
    del subscriptions
    with pytest.raises(RuntimeError, match="Subscriptions went out of scope"):
        handle.wait_for_ack(timeout=0.001)


@pytest.mark.parametrize("ack_rc", [MQTTReasonCode.Success, MQTTReasonCode.UnspecifiedError])
def test_subscriptions_wait_for_unsuback(
    ack_rc: MQTTReasonCode,
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test waiting for a SUBACK packet."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[ack_rc],
    )
    subscriptions.handle_suback(suback_packet)

    handle = subscriptions.unsubscribe("test/topic")
    assert handle is not None
    assert handle.wait_for_ack(timeout=0.001) is None

    # Simulate receiving an UNSUBACK packet
    unsuback_packet = MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[ack_rc],
    )
    subscriptions.handle_unsuback(unsuback_packet)

    assert handle.wait_for_ack(timeout=0.001) == unsuback_packet
    assert handle.ack == unsuback_packet

    mock_handlers.reset_mock()
    del mock_handlers
    mock_client.reset_mock()
    del mock_client
    mock_connection.reset_mock()
    del mock_connection
    del subscriptions
    with pytest.raises(RuntimeError, match="Subscriptions went out of scope"):
        handle.wait_for_ack(timeout=0.001)


@pytest.mark.parametrize("session_present", [True, False])
def test_subscriptions_subscribe_failure(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    session_present: bool
) -> None:
    """Test replaying SUBSCRIBE packets on reconnection after calling in a bad state."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    mock_connection.send.side_effect = InvalidStateError("TEST")
    handle = subscriptions.subscribe("test/topic", dummy_callback)

    assert handle is not None

    mock_connection.send.side_effect = None
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    if not session_present:
        # If the session was not present, we should send the SUBSCRIBE packet again.
        mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=1,
        ))


def test_subscriptions_unsubscribe_failure(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    mock_connection.send.side_effect = InvalidStateError("TEST")
    handle = subscriptions.unsubscribe("test/topic")

    assert handle is not None

    mock_connection.send.side_effect = None
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=True))

    mock_connection.send.assert_called_once_with(MQTTUnsubscribePacket(
        topics=["test/topic"],
        packet_id=1,
    ))
    mock_connection.reset_mock()


def test_subscriptions_unsubscribe_null(mock_handlers: MagicMock, mock_client: Mock, mock_connection: Mock) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle = subscriptions.unsubscribe("test/topic")
    assert handle is None


def test_subscriptions_unsubscribe_redundant(mock_handlers: MagicMock, mock_client: Mock, mock_connection: Mock) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    handle1 = subscriptions.unsubscribe("test/topic")
    handle2 = subscriptions.unsubscribe("test/topic")
    assert handle1 is not None
    assert handle1.failed
    subscriptions.handle_unsuback(MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))
    assert handle2 is not None
    assert handle2.ack is not None

    mock_connection.reset_mock()
    assert subscriptions.unsubscribe("test/topic") is None
    mock_connection.send.assert_not_called()


@pytest.mark.parametrize(("tf", "topic"), [
    ("test/topic", "test/topic"),
    ("test/+", "test/topic"),
    ("#", "test/topic"),
])
def test_subscriptions_handle_publish(
    tf: str,
    topic: str,
    mock_handlers: MagicMock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet."""
    mock_client = cast(Client, frozenset())
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd = []
    def callback(client: Client, packet: MQTTPublishPacket) -> None:
        recvd.append(packet)
    subscriptions.subscribe(tf, callback)

    publish_packet = MQTTPublishPacket(
        topic=topic,
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd == [publish_packet]

    del mock_client
    with pytest.raises(RuntimeError, match="Client went out of scope"):
        subscriptions.handle_publish(publish_packet)


def test_subscriptions_handle_publish_shared(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet with shared subscriptions."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd1 = []
    recvd2 = []
    def callback1(client: Client, packet: MQTTPublishPacket) -> None:
        recvd1.append(packet)
    def callback2(client: Client, packet: MQTTPublishPacket) -> None:
        recvd2.append(packet)

    subscriptions.subscribe("test/topic", callback1)
    subscriptions.subscribe("test/topic", callback2, share_name="test_share")

    publish_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd1 == [publish_packet]
    assert recvd2 == [publish_packet]


def test_subscriptions_handle_publish_sub_id(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet with subscription identifier."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd1 = []
    recvd2 = []
    def callback1(client: Client, packet: MQTTPublishPacket) -> None:
        recvd1.append(packet)
    def callback2(client: Client, packet: MQTTPublishPacket) -> None:
        recvd2.append(packet)

    subscriptions.subscribe("test/topic", callback1)
    subscriptions.subscribe("test/+", callback2, sub_id=1)

    publish_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        properties=MQTTPublishProps(SubscriptionIdentifier={1}),
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd1 == []
    assert recvd2 == [publish_packet]


def test_subscriptions_handle_publish_exception(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet with a broken callback."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd = []
    def callback1(client: Client, packet: MQTTPublishPacket) -> None:
        raise RuntimeError("TEST")
    def callback2(client: Client, packet: MQTTPublishPacket) -> None:
        recvd.append(packet)

    subscriptions.subscribe("test/topic", callback1)
    subscriptions.subscribe("test/+", callback2)

    publish_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd == [publish_packet]


def test_subscriptions_packet_id(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    # Test the private methods because the public interface is slow.
    for n in range(0xffff + 5):
        assert subscriptions._get_next_sub_packet_id() == (n % 0xffff) + 1  # noqa: SLF001

    for n in range(0xffff + 5):
        assert subscriptions._get_next_unsub_packet_id() == (n % 0xffff) + 1  # noqa: SLF001


def test_subscriptions_ack_unknown(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    with pytest.raises(MQTTError, match="Received SUBACK for unknown packet ID: 5") as excinfo:
        subscriptions.handle_suback(MQTTSubAckPacket(
            packet_id=5,
            reason_codes=[MQTTReasonCode.GrantedQoS2],
        ))
    assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError
    with pytest.raises(MQTTError, match="Received UNSUBACK for unknown packet ID: 6") as excinfo:
        subscriptions.handle_unsuback(MQTTUnsubAckPacket(
            packet_id=6,
            reason_codes=[MQTTReasonCode.Success],
        ))
    assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError


def test_subscriptions_handle_reset(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
) -> None:
    """Test failing handles when connection is reset."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle1 = subscriptions.subscribe("test/topic1", dummy_callback)
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))
    handle2 = subscriptions.subscribe("test/topic2", dummy_callback)
    handle3 = subscriptions.unsubscribe("test/topic1")
    subscriptions.handle_unsuback(MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.Success],
    ))
    handle4 = subscriptions.unsubscribe("test/topic2")

    subscriptions.handle_connack(MQTTConnAckPacket(
        session_present=False,
        reason_code=MQTTReasonCode.Success,
    ))

    assert handle1 is not None
    assert handle2 is not None
    assert handle3 is not None
    assert handle4 is not None
    assert not handle1.failed
    assert handle2.failed
    assert not handle3.failed
    assert handle4.failed


def test_subscriptions_slots(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))
    # Slots all the way down.
    assert not hasattr(subscriptions, "__dict__")
    assert hasattr(subscriptions, "__weakref__")
    assert all(hasattr(subscriptions, attr) for attr in subscriptions.__slots__)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
import threading
from typing import Callable, Final, Sequence, TYPE_CHECKING
import weakref

from .connection import Connection, InvalidStateError, MessageHandlers
from .error import MQTTError
from .logger import get_logger
from .mqtt_spec import MAX_PACKET_ID, MQTTReasonCode, MQTTQoS
from .packet import (
    MQTTPublishPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
)
from .property import MQTTSubscribeProps, MQTTUnsubscribeProps
from .topic_alias import InboundTopicAlias
from .topic_filter import validate_topic_filter, validate_share_name, join_share, match_topic_filter

if TYPE_CHECKING:
    from .client import Client

logger: Final = get_logger("subscriptions")

SubscribeCallback = Callable[["Client", MQTTPublishPacket], None]


class NoMatchingSubscriptionError(Exception):
    """Exception raised when no matching subscription is found when unsubscribing."""


class RetainPolicy(IntEnum):
    """Policy for broker sending retained messages upon a subscription.

    ALWAYS: Always send retained messages on subscription.

    ONCE: Only send retained messages on the first subscription to a topic.

    NEVER: Never send retained messages on subscription."""
    ALWAYS = 0
    ONCE = 1
    NEVER = 2


class _SubscriptionState(IntEnum):
    """Internal state of a subscription."""
    SUBSCRIBING = 0
    SUBSCRIBED = 1
    UNSUBSCRIBING = 2
    UNSUBSCRIBED = 3


class SubscribeHandle:
    """Represents a subscription to a topic filter with a callback."""
    __slots__ = ("__weakref__", "_subscriptions", "ack", "failed")

    ack: MQTTSubAckPacket | None
    failed: bool
    _subscriptions: weakref.ReferenceType[Subscriptions]

    def __init__(self, data: Subscription, subscriptions: weakref.ReferenceType[Subscriptions]) -> None:
        self._subscriptions = subscriptions
        self.ack = None
        self.failed = False

    def wait_for_ack(self, timeout: float | None = None) -> MQTTSubAckPacket | None:
        """Wait for the subscription acknowledgement.

        :raises RuntimeError: The Subscriptions object went out of scope.
        :return: The ack packet if acked, or None if the operation failed or timeout was reached."""
        subscriptions = self._subscriptions()
        if subscriptions is not None:
            return subscriptions.wait_for_suback(self, timeout=timeout)
        raise RuntimeError("Subscriptions went out of scope")


class UnsubscribeHandle:
    """Represents an unsubscribe request which was sent to the broker."""
    __slots__ = ("__weakref__", "_subscriptions", "ack", "failed")

    ack: MQTTUnsubAckPacket | None
    failed: bool
    _subscriptions: weakref.ReferenceType[Subscriptions]

    def __init__(self, subscriptions: weakref.ReferenceType[Subscriptions]) -> None:
        self.ack = None
        self.failed = False
        self._subscriptions = subscriptions

    def wait_for_ack(self, timeout: float | None = None) -> MQTTUnsubAckPacket | None:
        """Wait for the unsubscribe acknowledgement.

        :raises RuntimeError: The Subscriptions object went out of scope.
        :return: The ack packet if acked, or None if the operation failed or timeout was reached."""
        subscriptions = self._subscriptions()
        if subscriptions is not None:
            return subscriptions.wait_for_unsuback(self, timeout=timeout)
        raise RuntimeError("Subscriptions went out of scope")


@dataclass(match_args=True, slots=True)
class Subscription:
    """All the data about a request for subscription."""
    topic_filter: str
    callback: SubscribeCallback
    max_qos: MQTTQoS = MQTTQoS.Q2
    share_name: str | None = None
    no_local: bool = False
    retain_as_published: bool = False
    retain_policy: RetainPolicy = RetainPolicy.ALWAYS
    sub_id: int | None = None
    user_properties: tuple[tuple[str, str], ...] | None = None
    state: _SubscriptionState = field(default=_SubscriptionState.SUBSCRIBING)
    effective_filter: str = field(init=False, repr=False)

    def __post_init__(self) -> None:
        validate_topic_filter(self.topic_filter)
        if self.share_name is not None:
            validate_share_name(self.share_name)
        self.effective_filter = join_share(self.topic_filter, self.share_name)

    def render_sub(self) -> MQTTSubscribePacket:
        """Render the subscription into a SUBSCRIBE packet."""
        opts = self.max_qos.value | (self.retain_policy << 4) | (self.retain_as_published << 3) | (self.no_local << 2)
        props = MQTTSubscribeProps()
        if self.sub_id is not None:
            props.SubscriptionIdentifier = {self.sub_id}
        if self.user_properties is not None:
            props.UserProperty = self.user_properties
        return MQTTSubscribePacket(
            topics=[(self.effective_filter, opts)],
            properties=props,
        )

    def render_unsub(self) -> MQTTUnsubscribePacket:
        """Render the subscription into an UNSUBSCRIBE packet."""
        props = MQTTUnsubscribeProps()
        return MQTTUnsubscribePacket(
            topics=[self.effective_filter],
            properties=props,
        )


class Subscriptions:
    """Container for MQTT subscriptions and their callbacks."""
    __slots__ = (
        "__weakref__",
        "_client",
        "_cond",
        "_connection",
        "_inflight_sub_packet_ids",
        "_inflight_unsub_packet_ids",
        "_next_sub_packet_id",
        "_next_unsub_packet_id",
        "_sub_handles",
        "_subscriptions",
        "_topic_alias",
        "_unsub_handles",
    )

    def __init__(
        self,
        handlers: MessageHandlers,
        connection: Connection,
        client: weakref.ReferenceType[Client],
    ) -> None:
        self._connection = connection
        self._client = client
        self._cond = threading.Condition(connection.fsm.lock)
        self._topic_alias = InboundTopicAlias()
        self._subscriptions: dict[str, Subscription] = {}
        self._sub_handles: weakref.WeakValueDictionary[str, SubscribeHandle] = weakref.WeakValueDictionary()
        self._unsub_handles: weakref.WeakValueDictionary[str, UnsubscribeHandle] = weakref.WeakValueDictionary()
        self._inflight_sub_packet_ids: dict[int, str] = {}
        self._inflight_unsub_packet_ids: dict[int, str] = {}
        self._next_sub_packet_id = 1
        self._next_unsub_packet_id = 1

        handlers.register(MQTTSubAckPacket, self.handle_suback)
        handlers.register(MQTTUnsubAckPacket, self.handle_unsuback)

    def subscribe(
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: MQTTQoS = MQTTQoS.Q2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
    ) -> SubscribeHandle:
        """Add a subscription with a callback.

        Repeated calls to subscribe with the same combination of share_name and topic_filter will reconfigure
        the existing subscription with the new parameters, discarding the previous callback.

        :return: A SubscribeHandle which can be used to wait for ack."""
        sub = Subscription(
            topic_filter=topic_filter,
            callback=callback,
            max_qos=max_qos,
            share_name=share_name,
            no_local=no_local,
            retain_as_published=retain_as_published,
            retain_policy=retain_policy,
            sub_id=sub_id,
            user_properties=tuple(user_properties) if user_properties is not None else None,
        )
        with self._cond:
            self._subscriptions[sub.effective_filter] = sub
            if (existing := self._sub_handles.get(sub.effective_filter, None)) is not None and existing.ack is None and not existing.failed:
                existing.failed = True
                self._cond.notify_all()
            handle = SubscribeHandle(sub, weakref.ref(self))
            self._sub_handles[sub.effective_filter] = handle
            self._flush_packets()
            return handle

    def _flush_packets(self) -> None:
        """Try to send all pending SUBSCRIBE and UNSUBSCRIBE packets."""
        with self._cond:
            pending_subs = [
                sub for sub in self._subscriptions.values() if (
                    sub.state == _SubscriptionState.SUBSCRIBING and sub.effective_filter not in self._inflight_sub_packet_ids.values()
                )
            ]
            for sub in pending_subs:
                sub_packet = sub.render_sub()
                sub_packet.packet_id = self._get_next_sub_packet_id()
                self._inflight_sub_packet_ids[sub_packet.packet_id] = sub.effective_filter
                try:
                    self._connection.send(sub_packet)
                except InvalidStateError:
                    logger.debug("Connection closed, SUBSCRIBE not sent")
                    del self._inflight_sub_packet_ids[sub_packet.packet_id]
                    return

            pending_unsubs = [
                sub for sub in self._subscriptions.values() if (
                    sub.state == _SubscriptionState.UNSUBSCRIBING and sub.effective_filter not in self._inflight_unsub_packet_ids.values()
                )
            ]
            for sub in pending_unsubs:
                unsub_packet = sub.render_unsub()
                unsub_packet.packet_id = self._get_next_unsub_packet_id()
                self._inflight_unsub_packet_ids[unsub_packet.packet_id] = sub.effective_filter
                try:
                    self._connection.send(unsub_packet)
                except InvalidStateError:
                    logger.debug("Connection closed, UNSUBSCRIBE not sent")
                    del self._inflight_unsub_packet_ids[unsub_packet.packet_id]
                    return

    def _get_next_sub_packet_id(self) -> int:
        """Get the next SUBSCRIBE packet identifier."""
        sub_id = self._next_sub_packet_id
        self._next_sub_packet_id += 1
        if self._next_sub_packet_id > MAX_PACKET_ID:
            self._next_sub_packet_id = 1
        return sub_id

    def wait_for_suback(
        self,
        handle: SubscribeHandle,
        timeout: float | None = None,
    ) -> MQTTSubAckPacket | None:
        """Wait for a SUBACK for a SubscribeHandle.

        Returns the SUBACK packet if the subscription was sent to the broker and acknowledged.

        :return: None if the subscription was not sent to the broker or timeout was reached."""
        with self._cond:
            if self._cond.wait_for(lambda: handle.ack is not None or handle.failed, timeout=timeout):
                return handle.ack
            return None

    def unsubscribe(
        self,
        topic_filter: str,
        *,
        share_name: str | None = None,
    ) -> UnsubscribeHandle | None:
        """Unsubscribe from a topic filter.

        :return: UnsubscribeHandle if the topic was subscribed, otherwise None."""
        effective_filter = join_share(topic_filter, share_name)
        with self._cond:
            if effective_filter not in self._subscriptions:
                # Nothing to unsubscribe.
                return None
            sub = self._subscriptions[effective_filter]
            if sub.state == _SubscriptionState.UNSUBSCRIBED:
                return None
            sub.state = _SubscriptionState.UNSUBSCRIBING
            if (existing := self._unsub_handles.get(effective_filter, None)) is not None and existing.ack is None and not existing.failed:
                existing.failed = True
                self._cond.notify_all()
            handle = UnsubscribeHandle(weakref.ref(self))
            self._unsub_handles[effective_filter] = handle
            self._flush_packets()
            return handle

    def _get_next_unsub_packet_id(self) -> int:
        """Get the next UNSUBSCRIBE packet identifier."""
        unsub_id = self._next_unsub_packet_id
        self._next_unsub_packet_id += 1
        if self._next_unsub_packet_id > MAX_PACKET_ID:
            self._next_unsub_packet_id = 1
        return unsub_id

    def wait_for_unsuback(
        self,
        handle: UnsubscribeHandle,
        timeout: float | None = None,
    ) -> MQTTUnsubAckPacket | None:
        """Wait for a UNSUBACK for an UnsubscribeHandle.

        Returns the UNSUBACK packet if the subscription was sent to the broker and acknowledged.

        Returns None if the subscription was not sent to the broker or timeout was reached."""
        with self._cond:
            if self._cond.wait_for(lambda: handle.ack is not None or handle.failed, timeout=timeout):
                return handle.ack
            return None

    def handle_publish(self, packet: MQTTPublishPacket) -> None:
        """Handle incoming PUBLISH packets."""
        with self._cond:
            client = self._client()
            if client is None:
                raise RuntimeError("Client went out of scope")
            self._topic_alias.handle(packet)
            subs = [sub for sub in self._subscriptions.values() if match_topic_filter(sub.topic_filter, packet.topic)]
            if packet.properties.SubscriptionIdentifier is not None:
                subs = [sub for sub in subs if sub.sub_id in packet.properties.SubscriptionIdentifier]
            for sub in subs:
                try:
                    sub.callback(client, packet)
                except Exception:  # noqa: PERF203
                    logger.exception("Unhandled exception in subscription callback")

    def handle_suback(self, packet: MQTTSubAckPacket) -> None:
        """Handle incoming SUBACK packets."""
        if any(True for code in packet.reason_codes if code.is_error()):
            errs = [hex(code) for code in packet.reason_codes if code.is_error()]
            logger.error("Errors found in SUBACK return: %s", errs)
        with self._cond:
            effective_topic_filter = self._inflight_sub_packet_ids.pop(packet.packet_id, None)
            if effective_topic_filter is None:
                raise MQTTError(f"Received SUBACK for unknown packet ID: {packet.packet_id}", MQTTReasonCode.ProtocolError)
            sub = self._subscriptions.get(effective_topic_filter, None)
            if sub is not None and sub.state == _SubscriptionState.SUBSCRIBING:
                sub.state = _SubscriptionState.SUBSCRIBED
            handle = self._sub_handles.pop(effective_topic_filter, None)
            if handle is not None:
                handle.ack = packet
                self._cond.notify_all()

    def handle_unsuback(self, packet: MQTTUnsubAckPacket) -> None:
        """Handle incoming UNSUBACK packets."""
        if any(True for code in packet.reason_codes if code.is_error()):
            errs = [hex(code) for code in packet.reason_codes if code.is_error()]
            logger.error("Errors found in UNSUBACK return: %s", errs)
        with self._cond:
            effective_topic_filter = self._inflight_unsub_packet_ids.pop(packet.packet_id, None)
            if effective_topic_filter is None:
                raise MQTTError(f"Received UNSUBACK for unknown packet ID: {packet.packet_id}", MQTTReasonCode.ProtocolError)
            sub = self._subscriptions.get(effective_topic_filter, None)
            if sub is not None and sub.state == _SubscriptionState.UNSUBSCRIBING:
                sub.state = _SubscriptionState.UNSUBSCRIBED
            handle = self._unsub_handles.pop(effective_topic_filter, None)
            if handle is not None:
                handle.ack = packet
                self._cond.notify_all()

    def handle_connack(self, packet: MQTTConnAckPacket) -> None:
        """Handle incoming CONNACK packets."""
        with self._cond:
            self._reset_connection_states()
            self._flush_packets()

    def _reset_connection_states(self) -> None:
        """Reset all connection-level states."""
        with self._cond:
            self._next_sub_packet_id = 1
            self._next_unsub_packet_id = 1
            self._topic_alias.reset()
            should_notify = False
            for sub_handle in self._sub_handles.values():
                sub_handle.failed = True
                should_notify = True
            self._sub_handles.clear()
            for unsub_handle in self._unsub_handles.values():
                unsub_handle.failed = True
                should_notify = True
            self._unsub_handles.clear()
            self._inflight_sub_packet_ids.clear()
            self._inflight_unsub_packet_ids.clear()
            if should_notify:
                self._cond.notify_all()  # Should clear all waiting for handles.



from collections import deque
from dataclasses import dataclass, field
from threading import Condition
from typing import Final, Sequence
import weakref

from .base import Persistence, ReliablePublishHandle, RenderedPacket
from ..logger import get_logger
from ..mqtt_spec import MAX_PACKET_ID, MQTTQoS
from ..packet import MQTTPublishPacket, MQTTPubRelPacket
from ..property import MQTTPublishProps
from ..topic_alias import AliasPolicy

logger: Final = get_logger("persistence.in_memory")


@dataclass(match_args=True, slots=True)
class RetainedMessage:
    """Represents a qos>0 message in the session."""
    topic: str
    payload: bytes
    msg_id: int
    qos: MQTTQoS
    retain: bool
    properties: MQTTPublishProps
    dup: bool
    received: bool
    handle: weakref.ReferenceType[ReliablePublishHandle]
    alias_policy: AliasPolicy


@dataclass(slots=True)
class InMemoryPersistence(Persistence):
    """Store for retained messages in the session.

    This store is in memory only and is not persistent."""
    _client_id: str = field(default="", init=False)
    _next_msg_id: int = field(default=1, init=False)
    _messages: dict[int, RetainedMessage] = field(init=False, default_factory=dict)
    _packet_id_to_msg_id: dict[int, int] = field(init=False, default_factory=dict)
    _pending: deque[int] = field(init=False, default_factory=deque)
    _received: set[int] = field(init=False, default_factory=set)
    _cond: Condition = field(init=False, default_factory=Condition)

    def __len__(self) -> int:
        return len(self._messages)

    def add(
        self,
        topic: str,
        payload: bytes,
        qos: MQTTQoS,
        retain: bool,
        properties: MQTTPublishProps,
        alias_policy: AliasPolicy,
    ) -> ReliablePublishHandle:
        if alias_policy == AliasPolicy.ALWAYS:
            raise ValueError("AliasPolicy must not be ALWAYS for retained messages.")
        msg_id = self._next_msg_id
        self._next_msg_id += 1

        handle = ReliablePublishHandle(self._cond)
        message = RetainedMessage(
            topic=topic,
            payload=payload,
            msg_id=msg_id,
            qos=qos,
            retain=retain,
            properties=properties,
            dup=False,
            received=False,
            handle=weakref.ref(handle),
            alias_policy=alias_policy,
        )
        self._messages[msg_id] = message
        self._pending.append(msg_id)
        return handle

    def get(self, count: int) -> Sequence[int]:
        return [self._pending[i] for i in range(min(count, len(self._pending)))]

    def ack(self, packet_id: int) -> None:
        if packet_id not in self._packet_id_to_msg_id:
            raise ValueError(f"Unknown packet_id: {packet_id}")
        msg_id = self._packet_id_to_msg_id[packet_id]
        message = self._messages[msg_id]
        if message.qos == MQTTQoS.Q1 or message.received:
            handle = message.handle()
            if handle is not None:
                with self._cond:
                    handle.acked = True
                    self._cond.notify_all()
            del self._packet_id_to_msg_id[packet_id]
            del self._messages[msg_id]
        else:
            # Prioritize PUBREL over PUBLISH
            self._pending.appendleft(msg_id)
            message.received = True

    def check_rec(self, packet: MQTTPublishPacket) -> bool:
        if packet.qos != MQTTQoS.Q2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        if packet.packet_id in self._received:
            logger.debug("Received duplicate QoS 2 packet with ID %d", packet.packet_id)
            return False
        return True

    def set_rec(self, packet: MQTTPublishPacket) -> None:
        if packet.qos != MQTTQoS.Q2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        self._received.add(packet.packet_id)

    def rel(self, packet: MQTTPubRelPacket) -> None:
        self._received.remove(packet.packet_id)

    def render(self, msg_id: int) -> RenderedPacket:
        packet: MQTTPublishPacket | MQTTPubRelPacket
        msg = self._messages[msg_id]
        packet_id = msg.msg_id
        while packet_id > MAX_PACKET_ID:
            packet_id -= MAX_PACKET_ID
        if msg.received:
            alias_policy = AliasPolicy.NEVER
            packet = MQTTPubRelPacket(packet_id=packet_id)
        else:
            alias_policy = msg.alias_policy
            packet = MQTTPublishPacket(
                topic=msg.topic,
                payload=msg.payload,
                packet_id=packet_id,
                qos=msg.qos,
                retain=msg.retain,
                properties=msg.properties,
                dup=msg.dup,
            )
        if self._pending[0] != msg.msg_id:
            raise ValueError(f"Message {packet_id} is not next in queue.")
        self._pending.popleft()
        self._packet_id_to_msg_id[packet_id] = msg.msg_id
        return RenderedPacket(packet, alias_policy)

    def _reset_inflight(self) -> None:
        """Clear inflight status of all messages."""
        inflight = [i for i in self._messages.keys() if i not in self._pending]
        for msg_id in reversed(inflight):
            self._messages[msg_id].dup = True
            self._pending.appendleft(msg_id)

    def clear(self) -> None:
        with self._cond:
            self._messages.clear()
            self._pending.clear()
            self._packet_id_to_msg_id.clear()
            self._next_msg_id = 1

    def open(self, client_id: str, clear: bool = False) -> None:
        if clear or client_id != self._client_id:
            self.clear()
            self._client_id = client_id
        else:
            self._reset_inflight()



"""
Fake MQTT broker for testing purposes.
"""

from __future__ import annotations

from collections import deque
import socket
import socketserver
import threading
from typing import Callable, cast, Final
import uuid

from ohmqtt.connection.decoder import ClosedSocketError, IncrementalDecoder
from ohmqtt.logger import get_logger
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import (
    MQTTPacket,
    MQTTConnectPacket,
    MQTTConnAckPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
    MQTTPingReqPacket,
    MQTTPingRespPacket,
    MQTTDisconnectPacket,
)
from ohmqtt.property import MQTTConnAckProps
from ohmqtt.topic_filter import match_topic_filter, validate_topic_filter

logger: Final = get_logger("tests.util.fake_broker")


class FakeBrokerHandler(socketserver.BaseRequestHandler):
    server: FakeBrokerServer

    def setup(self) -> None:
        self.decoder = IncrementalDecoder()
        self.subscriptions: set[str] = set()

    def handle(self) -> None:
        self.server.sock = self.request
        try:
            while (packet := self.decoder.decode(self.request)) is not None:
                self._handle_packet(packet)
        except ClosedSocketError:
            pass

    def _handle_packet(self, packet: MQTTPacket) -> None:
        outbound: list[MQTTPacket] = []
        logger.info("FakeBroker <--- %s", packet)
        self.server.received.append(packet)

        handlers = {
            MQTTConnectPacket: self._handle_connect,
            MQTTSubscribePacket: self._handle_subscribe,
            MQTTUnsubscribePacket: self._handle_unsubscribe,
            MQTTPublishPacket: self._handle_publish,
            MQTTPubRecPacket: self._handle_pubrec,
            MQTTPubRelPacket: self._handle_pubrel,
            MQTTPingReqPacket: self._handle_pingreq,
            MQTTDisconnectPacket: self._handle_disconnect,
        }

        if type(packet) in handlers:
            handler = cast(Callable[[MQTTPacket], list[MQTTPacket]], handlers[type(packet)])
            outbound.extend(handler(packet))

        for pkt in outbound:
            logger.info("FakeBroker ---> %s", pkt)
            self.request.sendall(pkt.encode())

    def _handle_connect(self, packet: MQTTConnectPacket) -> list[MQTTPacket]:
        if packet.client_id:
            connack = MQTTConnAckPacket()
        else:
            client_id = f"auto-{uuid.uuid4()!s}"
            connack = MQTTConnAckPacket(
                properties=MQTTConnAckProps(AssignedClientIdentifier=client_id)
            )
        return [connack]

    def _handle_subscribe(self, packet: MQTTSubscribePacket) -> list[MQTTPacket]:
        for topic_filter, opts in packet.topics:
            validate_topic_filter(topic_filter)
            if opts & 0x04:
                # no_local is set, do not mirror.
                continue
            self.subscriptions.add(topic_filter)
        return [MQTTSubAckPacket(packet_id=packet.packet_id, reason_codes=[MQTTReasonCode.Success] * len(packet.topics))]

    def _handle_unsubscribe(self, packet: MQTTUnsubscribePacket) -> list[MQTTPacket]:
        for topic in packet.topics:
            validate_topic_filter(topic)
            self.subscriptions.discard(topic)
        return [MQTTUnsubAckPacket(packet_id=packet.packet_id, reason_codes=[MQTTReasonCode.Success] * len(packet.topics))]

    def _handle_publish(self, packet: MQTTPublishPacket) -> list[MQTTPacket]:
        outbound: list[MQTTPacket] = []
        if any(match_topic_filter(topic_filter, packet.topic) for topic_filter in self.subscriptions):
            outbound.append(packet)

        if packet.qos == MQTTQoS.Q1:
            outbound.append(MQTTPubAckPacket(packet_id=packet.packet_id))
        elif packet.qos == MQTTQoS.Q2:
            outbound.append(MQTTPubRecPacket(packet_id=packet.packet_id))

        return outbound

    def _handle_pubrec(self, packet: MQTTPubRecPacket) -> list[MQTTPacket]:
        return [MQTTPubRelPacket(packet_id=packet.packet_id)]

    def _handle_pubrel(self, packet: MQTTPubRelPacket) -> list[MQTTPacket]:
        return [MQTTPubCompPacket(packet_id=packet.packet_id)]

    def _handle_pingreq(self, packet: MQTTPingReqPacket) -> list[MQTTPacket]:
        return [MQTTPingRespPacket()]

    def _handle_disconnect(self, packet: MQTTDisconnectPacket) -> list[MQTTPacket]:
        self.request.close()
        return []

class FakeBrokerServer(socketserver.TCPServer):
    received: deque[MQTTPacket]
    sock: socket.socket | None

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.received = deque()
        self.sock = None


class FakeBroker(threading.Thread):
    def __init__(self) -> None:
        super().__init__(daemon=True)
        self.server = FakeBrokerServer(("localhost", 0), FakeBrokerHandler)
        self.port = self.server.server_address[1]

    def __enter__(self) -> FakeBroker:
        self.start()
        return self

    def __exit__(self, *args: object) -> None:
        # Shutting down the server deadlocks here, so pass and let it be dereferenced.
        pass

    @property
    def received(self) -> deque[MQTTPacket]:
        return self.server.received

    @property
    def sock(self) -> socket.socket | None:
        return self.server.sock

    def run(self) -> None:
        with self.server:
            self.server.serve_forever()

    def shutdown(self) -> None:
        self.server.shutdown()
        self.server.server_close()

    def send(self, packet: MQTTPacket) -> None:
        """Send a packet from the broker to the client."""
        if self.sock is None:
            raise RuntimeError("Broker-to-client socket is not initialized")
        logger.info("FakeBroker ---> %s", packet)
        self.sock.sendall(packet.encode())



from __future__ import annotations

import socket
import ssl
from typing import cast, Final, get_args

from .types import ConnectParams, StateData, StateEnvironment, ReceivablePacketT
from .fsm import FSM, FSMState
from .decoder import ClosedSocketError
from ..error import MQTTError
from ..logger import get_logger
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode
from ..packet import MQTTConnectPacket, MQTTConnAckPacket, MQTTDisconnectPacket, PING, PONG
from ..platform import AF_UNIX, HAS_AF_UNIX

logger: Final = get_logger("connection.states")


def _get_socket(family: socket.AddressFamily) -> socket.socket:
    """Get a socket object.

    This is patched in tests to use a mock or loopback socket."""
    return socket.socket(family, socket.SOCK_STREAM)


class ConnectingState(FSMState):
    """Connecting to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.keepalive.keepalive_interval = params.keepalive_interval
        state_data.timeout.interval = params.connect_timeout
        state_data.timeout.mark()
        state_data.connack = None
        state_data.disconnect_rc = None
        state_data.sock = _get_socket(params.address.family)
        if params.address.family in (socket.AF_INET, socket.AF_INET6):
            state_data.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, params.tcp_nodelay)
        state_data.decoder.reset()
        state_data.open_called = False
        with fsm.selector:
            fsm.selector.change_sock(state_data.sock)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("Connect timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            address = params.address
            if HAS_AF_UNIX and address.family == AF_UNIX:
                state_data.sock.connect(address.host)
            else:
                state_data.sock.connect((address.host, address.port))
        except ConnectionError as exc:
            logger.error("Failed to connect to broker: %s", exc)
            fsm.change_state(ClosedState)
            return True
        except (BlockingIOError, OSError):
            pass  # Either in progress or already connected, select will reveal which.

        with fsm.selector:
            timeout = state_data.timeout.get_timeout(max_wait)
            _, writable = fsm.selector.select(write=True, timeout=timeout)
        if writable:
            logger.debug("Socket connected to broker")
            if params.address.use_tls:
                fsm.change_state(TLSHandshakeState)
            else:
                fsm.change_state(MQTTHandshakeConnectState)
            state_data.sock.setblocking(False)
            return True
        return False


class TLSHandshakeState(FSMState):
    """Performing TLS handshake with the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        tls_context = params.tls_context if params.tls_context else ssl.create_default_context()
        state_data.sock = tls_context.wrap_socket(
            state_data.sock,
            server_hostname=params.tls_hostname if params.tls_hostname else params.address.host,
            do_handshake_on_connect=False,
        )
        with fsm.selector:
            fsm.selector.change_sock(state_data.sock)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("TLS handshake timeout")
            fsm.change_state(ClosedState)
            return True

        timeout = state_data.timeout.get_timeout(max_wait)
        sock = cast(ssl.SSLSocket, state_data.sock)
        try:
            logger.debug("trying TLS handshake")
            sock.do_handshake()
            fsm.change_state(MQTTHandshakeConnectState)
            return True
        except ssl.SSLWantReadError:
            logger.debug("TLS handshake wants read")
            with fsm.selector:
                fsm.selector.select(read=True, timeout=timeout)
        except ssl.SSLWantWriteError:
            logger.debug("TLS handshake wants write")
            with fsm.selector:
                fsm.selector.select(write=True, timeout=timeout)
        return False


class MQTTHandshakeConnectState(FSMState):
    """Sending MQTT CONNECT packet to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        connect_packet = MQTTConnectPacket(
            client_id=params.client_id,
            protocol_version=params.protocol_version,
            clean_start=params.clean_start,
            keep_alive=params.keepalive_interval,
            properties=params.connect_properties,
            will_topic=params.will_topic,
            will_payload=params.will_payload,
            will_qos=params.will_qos,
            will_retain=params.will_retain,
            will_props=params.will_properties,
            username=params.address.username,
            password=params.address.password.encode() if params.address.password else None,
        )
        logger.debug("---> %s", connect_packet)
        with fsm.selector:
            env.write_buffer.clear()
            env.write_buffer.extend(connect_packet.encode())

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("MQTT CONNECT keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            with fsm.selector:
                num_sent = state_data.sock.send(env.write_buffer)
                if num_sent == 0:
                    logger.error("MQTT CONNECT send returned 0 bytes, closing connection")
                    fsm.change_state(ClosedState)
                    return True
                if num_sent < len(env.write_buffer):
                    # Not all data was sent, wait for writable again.
                    logger.debug("Not all CONNECT data was sent, waiting for writable again: wrote: %d", num_sent)
                    del env.write_buffer[:num_sent]
                    return False
                env.write_buffer.clear()
            fsm.change_state(MQTTHandshakeConnAckState)
            return True
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            # The write was blocked, wait for the socket to be writable.
            if max_wait is None or max_wait > 0.0:
                with fsm.selector:
                    timeout = state_data.timeout.get_timeout(max_wait)
                    fsm.selector.select(write=True, timeout=timeout)
        return False


class MQTTHandshakeConnAckState(FSMState):
    """Receiving MQTT CONNACK packet from the broker."""
    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("MQTT CONNACK keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        want_read = False
        try:
            packet = state_data.decoder.decode(state_data.sock)
            if packet is None:
                want_read = True
        except ClosedSocketError:
            logger.exception("Socket was closed")
            fsm.change_state(ClosedState)
            return True

        if want_read:
            # Incomplete packet, wait for more data.
            with fsm.selector:
                timeout = state_data.timeout.get_timeout(max_wait)
                fsm.selector.select(read=True, timeout=timeout)
            return False

        if packet is not None and packet.packet_type == MQTTPacketType.CONNACK:
            packet = cast(MQTTConnAckPacket, packet)
            state_data.connack = packet
            if packet.properties.ServerKeepAlive is not None:
                state_data.keepalive.keepalive_interval = packet.properties.ServerKeepAlive
            fsm.change_state(ConnectedState)
            return True
        pt = packet.packet_type.name if packet is not None else "None"
        logger.error("Unexpected '%s' packet while waiting for CONNACK", pt)
        fsm.change_state(ClosedState)
        return True


class ConnectedState(FSMState):
    """Connected to the broker. Full duplex messaging in this state."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.keepalive.mark_init()
        with fsm.selector:
            env.write_buffer.clear()
        assert state_data.connack is not None, "Got to ConnectedState without a CONNACK"
        env.packet_callback(state_data.connack)
        state_data.open_called = True

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.keepalive.should_close():
            logger.error("Keepalive timeout, closing socket")
            fsm.change_state(ClosedState)
            return True

        if state_data.keepalive.should_send_ping():
            logger.debug("---> PING")
            with fsm.selector:
                env.write_buffer.extend(PING)
            state_data.keepalive.mark_ping()

        timeout = state_data.keepalive.get_next_timeout(max_wait)
        with fsm.selector:
            write_check = bool(env.write_buffer)
            readable, writable = fsm.selector.select(read=True, write=write_check, timeout=timeout)

        if writable:
            try:
                with fsm.selector:
                    sent = state_data.sock.send(env.write_buffer)
                    del env.write_buffer[:sent]
                state_data.keepalive.mark_send()
            except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                pass
            except (BrokenPipeError, ConnectionResetError) as exc:
                logger.error("MQTT connection was closed: %s", exc)
                fsm.change_state(ClosedState)
                return True

        if readable:
            want_read = True
            try:
                while want_read:  # Read all available packets.
                    want_read = cls.read_packet(fsm, state_data, env, params)
            except ClosedSocketError:
                logger.debug("Connection closed")
                fsm.change_state(ClosedState)
                return True
            except MQTTError as exc:
                logger.error("There was a problem with data from broker, closing connection: %s", exc)
                state_data.disconnect_rc = exc.reason_code
                fsm.change_state(ClosedState)
                return True

        return False

    @classmethod
    def read_packet(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> bool:
        """Called by the underlying SocketWrapper when the socket is ready to read.

        Incrementally reads and decodes a packet from the socket.
        Complete packets are passed up to the read callback.

        Returns False if an incomplete packet was read, True if a complete packet was read."""
        packet = state_data.decoder.decode(state_data.sock)
        if packet is None:
            # No complete packet available yet.
            return False

        # Ping requests and responses are handled at this layer.
        if packet.packet_type == MQTTPacketType.PINGRESP:
            logger.debug("<--- PONG")
            state_data.keepalive.mark_pong()
        elif packet.packet_type == MQTTPacketType.PINGREQ:
            logger.debug("<--- PING PONG --->")
            with fsm.selector:
                env.write_buffer.extend(PONG)
        elif packet.packet_type == MQTTPacketType.DISCONNECT:
            logger.debug("<--- %s", packet)
            logger.info("Broker sent DISCONNECT, closing connection")
            fsm.change_state(ClosingState)
            return True
        elif not isinstance(packet, get_args(ReceivablePacketT)):
            # To cast later, we must handle the exceptional cases at runtime.
            raise MQTTError("Unexpected packet type", reason_code=MQTTReasonCode.ProtocolError)
        else:
            # All other packets are passed to the read callback.
            env.packet_callback(cast(ReceivablePacketT, packet))
        return True


class ReconnectWaitState(FSMState):
    """Waiting to reconnect to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.timeout.interval = params.reconnect_delay
        state_data.timeout.mark()

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        # Here we repurpose the keepalive timer to wait for a reconnect.
        if state_data.timeout.exceeded():
            logger.debug("Reconnecting")
            fsm.change_state(ConnectingState)
            return True
        with fsm.cond:
            if fsm.state is not ReconnectWaitState:
                # The state has changed, don't wait.
                return True
            if max_wait is None or max_wait > 0.0:
                timeout = state_data.timeout.get_timeout(max_wait)
                fsm.cond.wait(timeout)
        return False


class ClosingState(FSMState):
    """Gracefully closing the connection.

    The socket will be shutdown for reading, but existing buffers will be flushed."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        if fsm.previous_state == ConnectedState:
            logger.debug("Gracefully closing connection")
            if state_data.disconnect_rc is None:
                state_data.disconnect_rc = MQTTReasonCode.NormalDisconnection
        else:
            logger.debug("Skipping ClosingState")
            fsm.change_state(ClosedState)
            return

        state_data.timeout.interval = params.connect_timeout
        state_data.timeout.mark()

        # Shutdown only the read side of the socket, so we can still send data.
        # We don't want to shutdown the socket if we're using TLS, it would cause the protocol to fail.
        if not params.address.use_tls:
            try:
                state_data.sock.shutdown(socket.SHUT_RD)
            except OSError as exc:
                logger.debug("Error while shutting down socket reading: %s", exc)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        # Wait for the socket to be writable.
        if state_data.timeout.exceeded():
            logger.error("ClosingState timed out")
            fsm.change_state(ClosedState)
            return True

        with fsm.selector:
            if not env.write_buffer:
                logger.debug("No more data to send, connection closed")
                fsm.change_state(ClosedState)
                return True
            timeout = state_data.timeout.get_timeout(max_wait)
            _, writable = fsm.selector.select(write=True, timeout=timeout)

            if writable:
                try:
                    sent = state_data.sock.send(env.write_buffer)
                    del env.write_buffer[:sent]
                    state_data.keepalive.mark_send()
                except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                    pass
                except BrokenPipeError:
                    logger.error("Socket lost while closing")
                    fsm.change_state(ClosedState)
                    return True
        return False


class ClosedState(FSMState):
    """Connection is closed.

    All buffers will be flushed and the socket will be closed after conditionally sending a DISCONNECT."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        if state_data.open_called:
            state_data.open_called = False
            with fsm.selector:
                if state_data.disconnect_rc is not None and not env.write_buffer:
                    disconnect_packet = MQTTDisconnectPacket(reason_code=state_data.disconnect_rc)
                    # Try to send a DISCONNECT packet, but no problem if we can't.
                    try:
                        state_data.sock.send(disconnect_packet.encode())
                        logger.debug("---> %s", disconnect_packet)
                    except OSError:
                        logger.debug("Failed to send DISCONNECT packet")
            if params.reconnect_delay > 0 and fsm.requested_state == ConnectingState:
                fsm.change_state(ReconnectWaitState)
        try:
            state_data.sock.shutdown(socket.SHUT_RDWR)
        except OSError as exc:
            logger.debug("Error while shutting down socket: %s", exc)
        try:
            state_data.sock.close()
        except OSError as exc:
            logger.debug("Error while closing socket: %s", exc)
        state_data.decoder.reset()
        with fsm.selector:
            env.write_buffer.clear()


class ShutdownState(FSMState):
    """The final state.

    All buffers will be flushed and the socket will be closed immediately on entry."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        # We can enter this state from any other state.
        # Free up as many resources as possible.
        try:
            state_data.sock.close()
        except OSError as exc:
            logger.debug("Error while closing socket: %s", exc)
        state_data.decoder.reset()
        with fsm.selector:
            env.write_buffer.clear()
            fsm.selector.close()


# Hook up transitions.
ConnectingState.transitions_to = (ClosingState, ClosedState, ShutdownState, TLSHandshakeState, MQTTHandshakeConnectState)
ConnectingState.can_request_from = (ClosedState, ReconnectWaitState)

TLSHandshakeState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnectState)

MQTTHandshakeConnectState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnAckState)

MQTTHandshakeConnAckState.transitions_to = (ClosingState, ClosedState, ShutdownState, ConnectedState)

ConnectedState.transitions_to = (ClosingState, ClosedState, ShutdownState)

ClosingState.transitions_to = (ClosedState, ShutdownState)
ClosingState.can_request_from = (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ConnectedState,
    ReconnectWaitState,
)

ClosedState.transitions_to = (ConnectingState, ShutdownState, ReconnectWaitState)

ReconnectWaitState.transitions_to = (ClosingState, ClosedState, ShutdownState, ConnectingState)

ShutdownState.can_request_from = (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ConnectedState,
    ReconnectWaitState,
    ClosingState,
    ClosedState,
)



import threading
import time

from ohmqtt.persistence.base import ReliablePublishHandle, UnreliablePublishHandle


def test_persistence_handles_unreliable_publish() -> None:
    """Test the UnreliablePublishHandle class."""
    handle = UnreliablePublishHandle()
    assert handle.is_acked() is False
    assert handle.wait_for_ack() is False
    assert handle.wait_for_ack(timeout=1) is False

    assert not hasattr(handle, "__dict__")
    assert all(hasattr(handle, attr) for attr in handle.__slots__)


def test_persistence_handles_reliable_publish() -> None:
    """Test the ReliablePublishHandle class."""
    cond = threading.Condition()
    handle = ReliablePublishHandle(cond)
    assert handle.is_acked() is False
    assert handle.wait_for_ack(timeout=0.001) is False

    def do_ack() -> None:
        time.sleep(0.1)
        with cond:
            handle.acked = True
            cond.notify_all()
    thread = threading.Thread(target=do_ack)
    thread.start()

    assert handle.wait_for_ack(timeout=1) is True
    assert handle.is_acked() is True
    thread.join()

    assert not hasattr(handle, "__dict__")
    assert all(hasattr(handle, attr) for attr in handle.__slots__)
